/*
	Copyright (C) 2000-2015 Claude SIMON (http://q37.info/contact/).

	This file is part of the Epeios framework.

	The Epeios framework is free software: you can redistribute it and/or
	modify it under the terms of the GNU Affero General Public License as
	published by the Free Software Foundation, either version 3 of the
	License, or (at your option) any later version.

	The Epeios framework is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
	Affero General Public License for more details.

	You should have received a copy of the GNU Affero General Public License
	along with the Epeios framework.  If not, see <http://www.gnu.org/licenses/>
*/

#define SCLRGSTRY__COMPILATION

#include "sclrgstry.h"

/******************************************************************************/
				  /* do not modify anything above this limit */
				  /*			  unless specified			 */
				  /*******************************************/

#include "cio.h"
#include "lcl.h"
#include "dir.h"
#include "fnm.h"
#include "tagsbs.h"

#include "sclerror.h"
#include "sclmisc.h"

using namespace sclrgstry;

#define REGISTRY_FILE_EXTENSION ".xcfg"

static rgstry::multi_level_registry Registry_;

static rgstry::level__ ConfigurationLevel_ = rgstry::UndefinedLevel;
static rgstry::level__ ProjectLevel_ = rgstry::UndefinedLevel;
static rgstry::level__ SetupLevel_ = rgstry::UndefinedLevel;
static rgstry::level__ ArgumentsLevel_ = rgstry::UndefinedLevel;

rgstry::entry___ sclrgstry::Parameters( ParametersTag );
rgstry::entry___ sclrgstry::Definitions( "Definitions" );
rgstry::entry___ sclrgstry::Locale( "Locale" );
rgstry::entry___ sclrgstry::Arguments( "Arguments" );

rgstry::entry___ sclrgstry::Language( "Language", Parameters );

static rgstry::entry___ PluginParameters_( "Plugins", sclrgstry::Parameters );
rgstry::entry___ sclrgstry::PluginParameter( RGSTRY_TAGGED_ENTRY( "Plugin", "target" ), PluginParameters_ );

static rgstry::entry___ PluginDefinitionsSection_( "PluginSection", sclrgstry::Definitions );
static rgstry::entry___ PluginDefinitions_( RGSTRY_TAGGED_ENTRY( "Plugins", "target" ), PluginDefinitionsSection_ );
static rgstry::entry___ PluginDefinition_( RGSTRY_TAGGED_ENTRY( "Plugin", "id" ), PluginDefinitions_ );
rgstry::entry___ sclrgstry::PluginFilename( "Filename", PluginDefinition_ );
rgstry::entry___ sclrgstry::PluginConfiguration( "Configuration", PluginDefinition_ );
rgstry::entry___ sclrgstry::PluginLocale( "Locale", PluginDefinition_ );


static rgstry::entry___ DefaultSetup_( "@DefaultSetup", Parameters );

static rgstry::entry___ Setup_( "@Setup", sclrgstry::Parameters );

registry_ &sclrgstry::GetCommonRegistry( void )
{
	return Registry_;
}

rgstry::level__ sclrgstry::GetConfigurationLevel( void )
{
	return ConfigurationLevel_;
}

rgstry::level__ sclrgstry::GetProjectLevel( void )
{
	return ProjectLevel_;
}

rgstry::level__ sclrgstry::GetSetupLevel( void )
{
	return SetupLevel_;
}

rgstry::level__ sclrgstry::GetArgumentsLevel( void )
{
	return ArgumentsLevel_;
}

const char *sclrgstry::GetLanguage_(
	const registry_ &Registry,
	TOL_CBUFFER___ &Buffer )
{
	sclrgstry::OGetValue( Registry, Language, Buffer );

	if ( Buffer == NULL )
		return sclmisc::GetBaseLanguage();
	else
		return Buffer;
}

static rgstry::status__ FillConfigurationRegistry_(
	xtf::extended_text_iflow__ &Flow,
	const char *Directory,
	const char *RootPath,
	rgstry::context___ &Context )
{
	return Registry_.Fill( ConfigurationLevel_, Flow, xpp::criterions___( str::string( Directory ) ), RootPath );
}

void sclrgstry::ReportBadOrNoValueForEntryErrorAndAbort( const rgstry::tentry__ &Entry )
{
ERRProlog
	str::string Path;
ERRBegin
	Path.Init();
	sclmisc::ReportAndAbort( SCLRGSTRY_NAME "_BadOrNoValueForEntry", Entry.GetPath( Path ) );
ERRErr
ERREnd
ERREpilog
}

static inline void ReportFileParsingErrorAndAbort_(
	const char *ErrorLabel,
	const rgstry::context___ &Context )
{
ERRProlog
	lcl::meaning Meaning;
	lcl::meaning MeaningBuffer;
ERRBegin
	Meaning.Init();
	Meaning.SetValue( ErrorLabel );

	MeaningBuffer.Init();
	rgstry::GetMeaning( Context, MeaningBuffer );

	Meaning.AddTag( MeaningBuffer );

	sclmisc::ReportAndAbort( Meaning );
ERRErr
ERREnd
ERREpilog
}

void sclrgstry::SetConfiguration( const rgstry::entry__ &Entry )
{
	Registry_.Erase( ConfigurationLevel_ );

	Registry_.Set( ConfigurationLevel_, Entry );
}

void sclrgstry::LoadConfiguration(
	xtf::extended_text_iflow__&Flow,
	const char *Directory,
	const char *RootPath )
{
ERRProlog
	rgstry::context___ Context;
ERRBegin
	Registry_.Erase( ConfigurationLevel_ );

	Context.Init();

	if ( FillConfigurationRegistry_( Flow, Directory, RootPath, Context ) != rgstry::sOK )
		ReportFileParsingErrorAndAbort_( SCLRGSTRY_NAME "_ConfigurationFileParsingError", Context );
ERRErr
ERREnd
ERREpilog
}

void sclrgstry::EraseProjectRegistry( void )
{
	Registry_.Erase( ProjectLevel_ );
}

#define PROJECT_ROOT_PATH	"Projects/Project[@target=\"%1\"]"

template <typename source> static void LoadProject_(
	source &Source,
	const char *Target,
	str::string_ &Id )
{
ERRProlog
	str::string Path;
	TOL_CBUFFER___ Buffer;
	rgstry::context___ Context;
ERRBegin
	Path.Init( PROJECT_ROOT_PATH );
	tagsbs::SubstituteShortTag( Path, 1, str::string( Target ), '%' );

	EraseProjectRegistry();

	Context.Init();
	if ( Registry_.Fill( ProjectLevel_, Source, xpp::criterions___(), Path.Convert( Buffer ), Context ) != rgstry::sOK )
		ReportFileParsingErrorAndAbort_( SCLRGSTRY_NAME "_ProjectFileParsingError", Context );

	Registry_.GetValue( ProjectLevel_, rgstry::entry___( "@Id" ), Id );
ERRErr
ERREnd
ERREpilog
}



void sclrgstry::LoadProject(
	flw::iflow__ &Flow,
	const char *Target,
	str::string_ &Id )
{
ERRProlog
	xtf::extended_text_iflow__ XFlow;
ERRBegin
	XFlow.Init( Flow, utf::f_Guess );

	LoadProject_( XFlow, Target, Id );
ERRErr
ERREnd
ERREpilog
}

void sclrgstry::LoadProject(
	const fnm::name___ &Filename,
	const char *Target,
	str::string_ &Id )
{
	LoadProject_( Filename, Target, Id );
}

void sclrgstry::EraseSetupRegistry( void )
{
	Registry_.Erase( SetupLevel_ );
}

static const str::string_ &GetSelectedSetupContent_(
	const str::string_ &SetupId,
	str::string_ &Content )
{
ERRProlog
	flx::E_STRING_OFLOW___ OFlow;
	txf::text_oflow__ TFlow;
	rgstry::row__ Row = E_NIL;
	rgstry::level__ Level = rgstry::UndefinedLevel;
	str::string SetupPath;
ERRBegin
	if ( SetupId.Amount() == 0  )
		ERRReturn;

	SetupPath.Init( "Setups/Setup[id=\"" );
	SetupPath.Append( SetupId );
	SetupPath.Append( "\"]" );

	OFlow.Init( Content );
	TFlow.Init( OFlow );

	Row = Registry_.Search( SetupPath, Level );

	if ( Row != E_NIL )
		Registry_.Dump( Level, Row, false, xml::oCompact, xml::e_None, TFlow );
ERRErr
ERREnd
ERREpilog
	return Content;
}

static void FillRegistryWithSetup_(
	registry_ &Registry,
	level__ Level,
	const str::string_ &RawSetupId )
{
ERRProlog
	str::string Setup;
	flx::E_STRING_IFLOW__ IFlow;
	xtf::extended_text_iflow__ XFlow;
	str::string EntryPath, SetupId;
ERRBegin
	SetupId.Init( RawSetupId );

	if ( SetupId.Amount() == 0 )
		Registry.GetValue( DefaultSetup_, SetupId );

	Setup.Init();

	Setup.Append('<' );
	sclrgstry::Parameters.GetPath( Setup );
	Setup.Append('>' );

	GetSelectedSetupContent_( SetupId, Setup );

	Setup.Append( "</" );
	sclrgstry::Parameters.GetPath( Setup );
	Setup.Append('>' );


	IFlow.Init( Setup );
	XFlow.Init( IFlow, utf::f_Default );
	Registry.Fill( Level, XFlow, xpp::criterions___(), NULL );
ERRErr
ERREnd
ERREpilog
}

void sclrgstry::FillSetupRegistry( const str::string_ &Id )
{
	FillRegistryWithSetup_( Registry_, SetupLevel_, Id );
}

void sclrgstry::FillSetupRegistry( void )
{
ERRProlog
	str::string Id;
ERRBegin
	Id.Init();
	OGetValue( Registry_, Setup_, Id );

	FillSetupRegistry( Id );
ERRErr
ERREnd
ERREpilog
}

void sclrgstry::ReportIfNoSetupId( void )
{
	if ( !Registry_.Exists( Setup_  ) )
		ReportBadOrNoValueForEntryErrorAndAbort( Setup_ );
}

bso::bool__ sclrgstry::BGetValue(
	const registry_ &Registry,
	const rgstry::tentry__ &Entry,
	str::string_ &Value )
{
	return Registry.GetValue( Entry, Value );
}

void sclrgstry::SetValue(
	registry_ &Registry,
	const str::string_ &Value,
	const rgstry::tentry__ &Entry )
{
	Registry.SetValue( Entry, Value );
}

void sclrgstry::SetValue(
	registry_ &Registry,
	const str::string_ &Path,
	const str::string_ &Value,
	sdr::row__ *Error )
{
	Registry.SetValue( Path, Value, Error );
}

bso::bool__ sclrgstry::GetValues(
	const registry_ &Registry,
	const rgstry::tentry__ &Entry,
	str::strings_ &Values )
{
	return Registry.GetValues( Entry, Values );
}

const str::string_ &sclrgstry::OGetValue(
	const registry_ &Registry,
	const rgstry::tentry__ &Entry,
	str::string_ &Value,
	bso::bool__ *Missing )
{
	if ( !BGetValue( Registry, Entry, Value ) )
		if ( Missing != NULL )
			*Missing = true;

	return Value;
}

const char *sclrgstry::OGetValue(
	const registry_ &Registry,
	const rgstry::tentry__ &Entry,
	TOL_CBUFFER___ &Buffer,
	bso::bool__ *Missing )
{
ERRProlog
	str::string Value;
	bso::bool__ LocalMissing = false;
ERRBegin
	Value.Init();

	OGetValue( Registry, Entry, Value, &LocalMissing );

	if ( LocalMissing ) {
		if ( Missing != NULL )
			*Missing = true;
	}
	else
		Value.Convert( Buffer );
ERRErr
ERREnd
ERREpilog
	return Buffer;
}

const str::string_ &sclrgstry::MGetValue(
	const registry_ &Registry,
	const rgstry::tentry__ &Entry,
	str::string_ &Value )
{
	if ( !BGetValue( Registry, Entry, Value ) )
		sclrgstry::ReportBadOrNoValueForEntryErrorAndAbort( Entry );

	return Value;
}

const char *sclrgstry::MGetValue(
	const registry_ &Registry,
	const rgstry::tentry__ &Entry,
	TOL_CBUFFER___ &Buffer )
{
ERRProlog
	str::string Value;
ERRBegin
	Value.Init();

	MGetValue( Registry, Entry, Value );

	Value.Convert( Buffer );
ERRErr
ERREnd
ERREpilog
	return Buffer;
}

static tol::extended_boolean__ GetBoolean_( const str::string_ &Value )
{
	tol::xbool__ Boolean = tol::xb_Undefined;
ERRProlog
	str::string Buffer;
ERRBegin
	Buffer.Init( Value );
	str::ToLower( Buffer );

	if ( ( Buffer == "true" ) || ( Buffer == "yes" ) )
		Boolean = tol::xbTrue;
	else if ( ( Buffer == "false" ) || ( Buffer == "no" ) )
		Boolean = tol::xbFalse;
ERRErr
ERREnd
ERREpilog
	return Boolean;
}

bso::bool__ sclrgstry::BGetBoolean(
	const registry_ &Registry,
	const rgstry::tentry__ &Entry,
	bso::bool__ DefaultValue )
{
	bso::bool__ &Result = DefaultValue;
ERRProlog
	str::string Value;
ERRBegin
	Value.Init();

	if ( BGetValue( Registry, Entry, Value ) ) {
		switch ( GetBoolean_( Value ) ) {
		case tol::xbFalse:
			Result = false;
			break;
		case tol::xbTrue:
			Result = true;
			break;
		case tol::xb_Undefined:
			sclrgstry::ReportBadOrNoValueForEntryErrorAndAbort( Entry );
			break;
		default:
			ERRFwk();
			break;
		}
	}
ERRErr
ERREnd
ERREpilog
	return Result;
}

bso::bool__ sclrgstry::MGetBoolean(
	const registry_ &Registry,
	const rgstry::tentry___ &Entry )
{
	bso::bool__ Result = false;
ERRProlog
	str::string Value;
ERRBegin
	Value.Init();

	switch ( GetBoolean_( MGetValue( Registry, Entry, Value ) ) ) {
	case tol::xbFalse:
		Result = false;
		break;
	case tol::xbTrue:
		Result = true;
		break;
	case tol::xb_Undefined:
		sclrgstry::ReportBadOrNoValueForEntryErrorAndAbort( Entry );
		break;
	default:
		ERRFwk();
		break;
	}
ERRErr
ERREnd
ERREpilog
	return Result;
}


template <typename t> static bso::bool__ GetUnsignedNumber_(
	const registry_ &Registry,
	const rgstry::tentry__ &Entry,
	t Limit,
	t &Value )
{
	bso::bool__ Present = false;
ERRProlog
	str::string RawValue;
	sdr::row__ Error = E_NIL;
ERRBegin
	RawValue.Init();

	if ( !( Present = BGetValue( Registry, Entry, RawValue ) ) )
		ERRReturn;

	RawValue.ToNumber( Value, Limit, &Error );

	if ( Error != E_NIL )
		sclrgstry::ReportBadOrNoValueForEntryErrorAndAbort( Entry );
ERRErr
ERREnd
ERREpilog
	return Present;
}

template <typename t> static bso::bool__ GetSignedNumber_(
	const registry_ &Registry,
	const rgstry::tentry__ &Entry,
	t LowerLimit,
	t UpperLimit,
	t &Value )
{
	bso::bool__ Present = false;
ERRProlog
	str::string RawValue;
	sdr::row__ Error = E_NIL;
ERRBegin
	RawValue.Init();

	if ( !( Present = BGetValue( Registry, Entry, RawValue ) ) )
		ERRReturn;

	RawValue.ToNumber( Value, UpperLimit, LowerLimit, &Error );

	if ( Error != E_NIL )
		sclrgstry::ReportBadOrNoValueForEntryErrorAndAbort( Entry );
ERRErr
ERREnd
ERREpilog
	return Present;
}

#define UN( name, type )\
type sclrgstry::MGet##name(\
	const registry_ &Registry,\
	const rgstry::tentry__ &Entry,\
	type Limit  )\
{\
	type Value;\
	\
	if ( !GetUnsignedNumber_( Registry, Entry, Limit, Value ) )\
		sclrgstry::ReportBadOrNoValueForEntryErrorAndAbort( Entry );\
	\
	return Value;\
}\
type sclrgstry::OGet##name(\
	const registry_ &Registry,\
	const rgstry::tentry__ &Entry,\
	type DefaultValue,\
	type Limit )\
{\
	type Value;\
	\
	if ( !GetUnsignedNumber_( Registry, Entry, Limit, Value ) )\
		Value = DefaultValue;\
	\
	return Value;\
}


UN( UInt, bso::uint__ )
#ifdef BSO__64BITS_ENABLED
UN( U64, bso::u64__ )
#endif
UN( U32, bso::u32__ )
UN( U16, bso::u16__ )
UN( U8, bso::u8__ )

#define SN( name, type )\
type sclrgstry::MGet##name(\
	const registry_ &Registry,\
	const rgstry::tentry__ &Entry,\
	type Min,\
	type Max)\
{\
	type Value;\
	\
	if ( !GetSignedNumber_( Registry, Entry, Min, Max, Value ) )\
		sclrgstry::ReportBadOrNoValueForEntryErrorAndAbort( Entry );\
	\
	return Value;\
}\
type sclrgstry::OGet##name(\
	const registry_ &Registry,\
	const rgstry::tentry__ &Entry,\
	type DefaultValue,\
	type Min,\
	type Max )\
{\
	type Value;\
	\
	if ( !GetSignedNumber_( Registry, Entry, Min, Max, Value ) )\
		Value = DefaultValue;\
	\
	return Value;\
}

SN( SInt, bso::sint__ )
#ifdef BSO__64BITS_ENABLED
SN( S64, bso::s64__ )
#endif
SN( S32, bso::s32__ )
SN( S16, bso::s16__ )
SN( S8, bso::s8__ )

Q37_GCTOR( sclrgstry )
{
	Registry_.Init();

	ConfigurationLevel_ = Registry_.Create();
	ProjectLevel_ = Registry_.Create();
	SetupLevel_ = Registry_.Create();
	ArgumentsLevel_ = Registry_.CreateEmbedded();
}
